#define ADDR__BASE  10          // начало области EEPROM для этой программы   
#define ADDR__TIMER_START         (ADDR__BASE+0) // для EEPROM
#define ADDR__TIMER_TOFLAG_COUNT  (ADDR__BASE+1)
#define ADDR__PULSE_LENGTH        (ADDR__BASE+2)
#define ADDR__RECALC_LEN_0        (ADDR__BASE+3) // оно двухбайтовое
#define ADDR__RECALC_LEN_1        (ADDR__BASE+4)

#define PIN_OUT      13
#define OSCILL_POINTS 40          //  сколько делать измерений
#define OSCILL_POINTS_PAUSE 1     // пауза между измерениями OSCILL_POINTS_PAUSE * OSCILL_POINTS < 256!!!
#define MAX_TIMER_START 250    // чаще не надо

#define KEY_RIGHT   0             // LCD
#define KEY_UP      1
#define KEY_DOWN    2
#define KEY_LEFT    3
#define KEY_SELECT  4
#define NUM_KEYS    5

#define CMD__RECALC_LEN         0
#define CMD__PULSE_LENGTH       1
#define CMD__TIMER_START     2
#define CMD__TIMER_TOFLAG_COUNT 3
#define CMD__STATUS             4
#define COMMANDS_COUNT          (CMD__STATUS+1)

#define FASTADC 1 // defines for setting and clearing register bits
#ifndef cbi       // при FASTADC 1 analogRead работает около 20мкс, в обычном состоянии около 100
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#include <EEPROM.h>
#include <LCD4Bit_mod.h> 
// для экранчика, если его нет, вычищать не нужно, просто все что с нис связано не будет работать
char commandNames[COMMANDS_COUNT][13]={
"RECALC COUNT",
"PULSE LENGTH",
"TIMER START",
"TOFLAG COUNT", 
"STATUS"};
char currCommand=2;
int adc_key_val[5] ={30, 150, 360, 535, 760 };
int adc_key_in, key=-1, oldkey=-1;
bool LCD_WRITING=false, statusPart=false;


unsigned int recalcCount=0,    // счётчик полного прохода OSCILL_POINTS измерений
             RECALC_LEN=100; // < 65536 раз во сколько recalcCount усредняются и отправляется данные
volatile boolean TIMER_FLAG         = false; // флаг поднимаемый прерыванием таймера. Сигнализирует, что пора слать следующий импульс
volatile uint8_t TIMER_COUNTER      = 0;   // счетчик срабатываний прерывания таймера
volatile uint8_t TIMER_START     = 10;  // начальное значение для таймера. Считать будет от TIMER_START до 255, на 255+1=0 сработает прерывание
volatile uint8_t TIMER_TOFLAG_COUNT = 20;  // <=255 сколько раз должна быть вызвана процедура прерывания до поднятия флага
// прикинуть время между флагами можно примерно так
// таймер срабатывает каждые (256-TIMER_START)*( 16MHz/8 = 2Mhz or 0.5us) us
// при этом TIMER_COUNTER увеличивается на 1
// когда TIMER_COUNTER >= TIMER_FLAG_TIME, взводится флаг

volatile uint8_t *REG_8_12_13;// адрес второго порта
volatile uint8_t MASK_12;     // маска 12 ноги
volatile uint8_t MASK_13;     // маска 13 ноги

uint8_t PULSE_LENGTH  = 200; // длительность импульса в микросекундах

unsigned long oscillMeasureTime=0;
unsigned long akbVoltage=0;  // переменная для измерения напряжения на АКБ непосредственно перед импульсом 
unsigned long oscillPointsAcc[ OSCILL_POINTS ]; // накопители для данных "осциллографа"
uint8_t analogReadNum=0;     // индекс текущего измерения

String sep=":";
String outStr;
String cmd="";

LCD4Bit_mod lcd = LCD4Bit_mod(2); 


ISR( TIMER2_OVF_vect ) { // процедура прерывания
    TIMER_COUNTER++;
    if( TIMER_COUNTER > TIMER_TOFLAG_COUNT ){
        TIMER_FLAG   =true;
        TIMER_COUNTER=0;
    }
    TCNT2=TIMER_START; 
}

void timer2Start( unsigned char tValue ){ // инициализация и запуск таймера
    //Timer2 Settings: Timer Prescaler /8, mode 0
    //Timmer clock = 16MHz/8 = 2Mhz or 0.5us
    //The /8 prescale gives us a good range to work with so we just hard code this for now.
    //load the timer for its first cycle
    TCNT2=tValue; 
    TCCR2A = 0;
    TCCR2B = 0<<CS22 | 1<<CS21 | 0<<CS20; 
    //Timer2 Overflow Interrupt Enable   
    TIMSK2 = 1<<TOIE2;
    //load the timer for its first cycle
    TCNT2=tValue; 
}



void serialStrVal( String s, int i ){
    Serial.print( s );
    Serial.println( i );
}
void serialStrVal( String s, unsigned int i ){
    Serial.print( s );
    Serial.println( i );
}
void serialStrVal( String s, float i ){
    Serial.print( s );
    Serial.println( i );
}
void serialStrVal( String s, char i ){
    Serial.print( s );
    Serial.println( i );
}


void do_RECALC_LEN( unsigned int val ){ // сменить частоту переподсчётов
    RECALC_LEN = min( 0xffff, max( 10, val ));
    EEPROM.write ( ADDR__RECALC_LEN_0, (RECALC_LEN >> 8));
    EEPROM.write ( ADDR__RECALC_LEN_1, (RECALC_LEN&0xFF));
    serialStrVal("m:New RECALC_LEN set to ", RECALC_LEN );
}

void do_TIMER_START( unsigned char val ){ // сменить длительность таймера
   TIMER_START = min( MAX_TIMER_START, max( 1, val ) );
   EEPROM.write ( ADDR__TIMER_START, TIMER_START);
   serialStrVal("m:TIMER_START set to ", TIMER_START );
}
void do_TIMER_TOFLAG_COUNT( unsigned char val ){ // сменить счетчик до флага
   TIMER_TOFLAG_COUNT = max( 1, val );
   EEPROM.write ( ADDR__TIMER_TOFLAG_COUNT, TIMER_TOFLAG_COUNT );
   serialStrVal("m:TIMER_TOFLAG_COUNT  set to ", TIMER_TOFLAG_COUNT  );
}

void do_PULSE_LENGTH( int val ){ // сменить длительность импульса
   PULSE_LENGTH = min( 255, max( 0, val ));
   EEPROM.write ( ADDR__PULSE_LENGTH, PULSE_LENGTH );
   serialStrVal("m:PULSE_LENGTH set to ", PULSE_LENGTH );
}

void do_STATUS(){ // показать статус
   Serial.println("");
   serialStrVal("\n\nm:(p=) PULSE_LENGTH=", PULSE_LENGTH );
   serialStrVal("m:(t=) TIMER_START=", TIMER_START );
   serialStrVal("m:(f=) TIMER_FLAG_COUNT=", TIMER_TOFLAG_COUNT );
   serialStrVal("m:(r=) RECALC_LEN=", RECALC_LEN );
   Serial.println("m:(s=) print status" );
   Serial.println("");
}


void doCmd(){
   if( cmd.length()>0 ){
      if( cmd.length()>=140 ){
	 Serial.println( "\nm:Line too long!" );
      }
      cmd.toLowerCase();
      char mnemo =cmd.charAt( 0 );
      int possInt=cmd.substring( 2 ).toInt();
      Serial.print( "c:");
      Serial.println( cmd );
      switch( mnemo ){
        case 'r': do_RECALC_LEN( possInt ); break;
        case 't': do_TIMER_START( possInt ); break;
        case 'p': do_PULSE_LENGTH( possInt ); break;
        case 'f': do_TIMER_TOFLAG_COUNT( possInt ); break;
        case 's': do_STATUS(); break;
      default:
        Serial.println( "Command not found" );
        do_STATUS();
      }
      cmd="";
      lcdStatus( true );
   }
}

void readSerial(){
   while( Serial.available() ){
      char ch=Serial.read();
      if( ch == 10 || ch == 13 || cmd.length()>=140 ){
	 doCmd();
      } else {
	 cmd+=ch ;
      }
   }
}

void loadEEPROMvalues(){
    TIMER_START     = EEPROM.read ( ADDR__TIMER_START);
    TIMER_TOFLAG_COUNT = EEPROM.read ( ADDR__TIMER_TOFLAG_COUNT );
    PULSE_LENGTH       = EEPROM.read ( ADDR__PULSE_LENGTH );    
    RECALC_LEN	       = EEPROM.read ( ADDR__RECALC_LEN_0);
    RECALC_LEN	<<= 8;
    RECALC_LEN	      += EEPROM.read ( ADDR__RECALC_LEN_1);
    // если эта область еще не использовалась, во всех ячейках сиди 0xFF
    if( (TIMER_START & TIMER_TOFLAG_COUNT & PULSE_LENGTH & RECALC_LEN ) == 0xFF ) saveEEPROMdefaults();
        
}
void saveEEPROMdefaults(){
    TIMER_START = 10;
    EEPROM.write ( ADDR__TIMER_START, TIMER_START );
    TIMER_TOFLAG_COUNT = 10;
    EEPROM.write ( ADDR__TIMER_TOFLAG_COUNT, TIMER_TOFLAG_COUNT );
    PULSE_LENGTH   = 100;
    EEPROM.write ( ADDR__PULSE_LENGTH, PULSE_LENGTH );    
    RECALC_LEN	  = 100;
    EEPROM.write ( ADDR__RECALC_LEN_0, (RECALC_LEN>>8) );
    EEPROM.write ( ADDR__RECALC_LEN_1, (RECALC_LEN&0xFF));
}


void lcdStatus( boolean dontChangePart ){
    if( !dontChangePart ) statusPart = !statusPart;
    String s="";
    if( statusPart  ){
        s+="p"+String(PULSE_LENGTH);
        s+=" t"+String(TIMER_START);
        s+=" f"+String(TIMER_TOFLAG_COUNT);
    }else{
        s+=" r"+String(RECALC_LEN);
    }
    
    while( s.length()<16 ) s+=" ";
    char c[16];
    s.toCharArray(c, 16);
    
    LCD_WRITING=true;
    lcd.cursorTo(2,0);
    lcd.printIn(c);
    LCD_WRITING=false;
}
void lcdPrintCmd(){
    String s=commandNames[ (int)currCommand ]; // в s  название текущей команды
    int cmdVal=0;
    if(CMD__STATUS != currCommand){ // у команды status значения нет
        switch( currCommand ){
            case CMD__RECALC_LEN        : cmdVal=RECALC_LEN; break;
            case CMD__TIMER_START       : cmdVal=TIMER_START; break;
            case CMD__TIMER_TOFLAG_COUNT: cmdVal=TIMER_TOFLAG_COUNT; break;
            case CMD__PULSE_LENGTH      : cmdVal=PULSE_LENGTH; break;
            case CMD__STATUS     : ;
        }
        s+=(cmdVal<100?" ":"")+String(cmdVal); // приклеиваем значение к названию текущей команды
    }
    while( s.length()<16 ) s+=" "; // дополняем строку
    char c[16];
    s.toCharArray( c, 16 );
    LCD_WRITING=true;
    lcd.cursorTo(1,0);
    lcd.printIn( c ); 
    LCD_WRITING=false;
}

void cmdAdd(int inc){
    currCommand+=inc;
    if(currCommand>=COMMANDS_COUNT){
        currCommand=0;
    }
    if(currCommand<0){
        currCommand=COMMANDS_COUNT-1;
    }
    lcdPrintCmd();
}

void cmdValueAdd( int inc ){
    switch( currCommand ){
        case CMD__RECALC_LEN           : do_RECALC_LEN( RECALC_LEN + inc*10); break;
        case CMD__TIMER_START       : do_TIMER_START( TIMER_START + inc ); break;
        case CMD__PULSE_LENGTH         : do_PULSE_LENGTH( PULSE_LENGTH + inc ); break;
        case CMD__TIMER_TOFLAG_COUNT   : do_TIMER_TOFLAG_COUNT( TIMER_TOFLAG_COUNT + inc ); break;
        case CMD__STATUS               : do_STATUS(); break;
    }
    lcdPrintCmd();
    lcdStatus( true );
}

int  get_key( int input){ // Convert ADC value to key number
	int k;
	for (k = 0; k < NUM_KEYS; k++){
		if (input < adc_key_val[k]) return k;
	}
    if (k >= NUM_KEYS) k = -1;     // No valid key pressed
    return k;
}
void doKeys(){
    adc_key_in = analogRead(0);        // read the value from the sensor  
    key = get_key(adc_key_in);		   // convert into key press

    if (key != oldkey){				   // if keypress is detected
        delay(50);		                   // wait for debounce time
        adc_key_in = analogRead(0);    // read the value from the sensor  
        key = get_key(adc_key_in);		   // convert into key press
        if (key != oldkey){
            oldkey = key;
            switch( key ){
                case KEY_RIGHT: cmdValueAdd( 1); break;
                case KEY_UP   : cmdAdd(1); break;
                case KEY_DOWN : cmdAdd(-1); break;
                case KEY_LEFT : cmdValueAdd(-1); break;
                case KEY_SELECT  : lcdStatus( false ); return;
                default:
                    return;
            }
            lcdStatus( true );
        }
    }
}



// запуск АЦП, через какоето время данные дождёмся и считаем
void myAnalogReadStart(uint8_t pin){
	if (pin >= 14) pin -= 14; // allow for channel or pin numbers
    volatile uint8_t _ADMUX;
    _ADMUX=(DEFAULT << 6) | (pin & 0x07);
	ADMUX = (DEFAULT << 6) | (pin & 0x07);
	sbi(ADCSRA, ADSC);
}

// ждем готовности и считываем данные ранее запущенного процесса АЦП
int myAnalogReadGet(){
	uint8_t low, high;
	// ADSC is cleared when the conversion finishes
	while (bit_is_set(ADCSRA, ADSC));
	low  = ADCL;
	high = ADCH;
	return (high << 8) | low; // combine the two bytes
} 



void setup(){
    loadEEPROMvalues();
    for( int i=0; i < OSCILL_POINTS; i++ ) oscillPointsAcc[i]=0;
    REG_8_12_13	= portOutputRegister( digitalPinToPort( 8 ) );
    MASK_12 	= digitalPinToBitMask( 12 );
    MASK_13 	= digitalPinToBitMask( 13 );

    Serial.begin( 57600 );
    Serial.println( "" );
    Serial.println( "#graph,0,1,p,16,AKB,0,0,255" );
    Serial.print( "#oscill,"); Serial.print( OSCILL_POINTS);  Serial.println(",16,AKB Pulse,255,0,0" );
    Serial.println( "#SysMessagesDone" );
    Serial.println( "m:Starting" );

    #if FASTADC // set prescale to 16
      sbi(ADCSRA,ADPS2) ;
      cbi(ADCSRA,ADPS1) ;
      cbi(ADCSRA,ADPS0) ;
    #endif

    LCD_WRITING=true;
        lcd.init();
        lcd.clear();
        cmdAdd( 0 );
        lcdStatus( true );
    LCD_WRITING=false;
   
    pinMode( PIN_OUT,   OUTPUT  );
    timer2Start( TIMER_START );
    do_STATUS();
    
}
volatile uint8_t myR18;


void loop() {
    while( !TIMER_FLAG || LCD_WRITING ){}; // ждем флага таймера
        
    {   // это один импульс и чтение осциллограммы после него
        // осциллограмма состоит из OSCILL_POINTS точек пауза между ними OSCILL_POINTS_PAUSE * (сколько там тактов нагенерит компилятор )
        // в лучшем случае компилятор сделает цикл со временем выполнения 2 такта = 250nS
        // т.о. если взять OSCILL_POINTS_PAUSE=4, то замеры будем производить через 1мкс
        akbVoltage+= analogRead(A1);      // меряем напряжение на АКБ перед импульсом, эта процедура займет 20мкс
        myR18= analogReadNum*OSCILL_POINTS_PAUSE;

        ADMUX = 0x41; // A1
        asm("ldi r24,lo8(65)");
        asm("sts 124,r24");
        *REG_8_12_13 |= MASK_13;     // не ассемблер потому как дальше delayMicroseconds, один хрен поправит всё ))
        asm("lds r24,PULSE_LENGTH");
        asm("ldi r25,lo8(0)");
        asm("call delayMicroseconds");
        if( analogReadNum == 0 ){    // такой финт, чтобы нулевой замер происходил прямо перед снятием импульса
                                     // это даёт нам возможность оценить насколько хорошо усвоился предыдущий импульс  
            sbi(ADCSRA, ADSC);       //  == myAnalogReadStart(A1), только без ожидания считывания
           *REG_8_12_13 &= ~MASK_13; // снятие импульса
            oscillPointsAcc[ 0 ]+=myAnalogReadGet(); // ожидаем и считываем данные АЦП
        }else{
           *REG_8_12_13 &= ~MASK_13;
          asm("lds r18,myR18");     // 2 такта for( i=R18; i>0; i-- ) // ожидаем время начала чтения для текущего analogReadNum
          asm("tst r18");           // 1 такт
          asm("breq .my_R18_zero"); // 1 такт , 2 если OSCILL_POINTS_PAUSE==0
                                    // получается, что минимальная пауза между замерами 5+1nop+2 на sbi = 8*0.0625 = 0.5мкс
                                    // а каждая единица OSCILL_POINTS_PAUSE добавляет еще по 0,5
                                    // если нужно мерять реже, нужно увеличивать длинну цикла
      asm(".my_R18_nonzero:");      // 1 цикл без NOPов 3 такта=62.5нс, => 1 OSCILL_POINTS_PAUSE = 0,5мкс
  
          asm("nop\n nop\n nop\n nop\n nop\n ");    // 5*62,5ns + 3*62.5 конец цикла всего 0.5мкс
          
          asm("dec r18"); 	                        // 1 такт
          asm("brne .my_R18_nonzero");              // 2 такта если переход есть, 1 если нету
      asm(".my_R18_zero:");
          asm("nop");                               // перехода нет, добавляем недостающий такт
  
          sbi(ADCSRA, ADSC); //        myAnalogReadStart(A1);
          oscillPointsAcc[ analogReadNum ]+=myAnalogReadGet();
        }

        analogReadNum++;
        if( analogReadNum >= OSCILL_POINTS ){ // провели считывание всех точек в пакете "осциллографа"?
            analogReadNum=0; // опять начнем с 0
            recalcCount++;     // увеличиваем счётчик усреднений
        }
        
        if( recalcCount >= RECALC_LEN  ){         
            // считаем время измерений
            // длинна цикла * OSCILL_POINTS_PAUSE*analogReadNum мкс

            outStr = String("o:")+ // метка строки осциллографа
                +((OSCILL_POINTS_PAUSE*OSCILL_POINTS)>>1);
                
            for( int i=0; i<OSCILL_POINTS; i++ ){ 
                outStr +=  sep+String( int( oscillPointsAcc[i] / recalcCount ));
                oscillPointsAcc[i]=0;
            }
            Serial.println( outStr ); // печать строки осциллографа
            
            outStr = String("g:")+ int( akbVoltage/OSCILL_POINTS/recalcCount );
            akbVoltage=0;
            Serial.println( outStr ); // печать строки аккумулятора
            recalcCount=0;
        }
    }
    readSerial();
    doKeys();
    TIMER_FLAG=false;
}
/* потом проверить
Set the ADLAR bit in ADMUX register and read just ADCH
For example,
    uint8_t readADC8(uint8_t channel)
    {
        ADMUX = (channel & 0x1F) | _BV(ADLAR);
        ADCSRA |= _BV(ADSC); // in atmega16 RMW operation actually _clears_ ADIF if it was set before
        while( !(ADCSRA & _BV(ADIF)) ) ;
        return ADCH;
    }
*/

